import argparse
import sys
from truffle.common import get_logger

logger = get_logger()

class ColorArgParser(argparse.ArgumentParser):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
    
    def _print_message(self, message, file=None):  
        if file == sys.stderr:
            logger.error(message)
        else:
            logger.plain_info(message)


from typing import List, Dict, Any

from .runtime import Runtime, tool, args, group, app, flags, HOST
from .common import get_logger



from .api import *
from .types import *

import logging 

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
)
logger = get_logger()

env : api.TGlobals = api.TGlobals()

def run(class_instance: Any) -> Any:
    logger.info(f"Building and running Truffle app: {class_instance.__class__.__name__}")
    rt = Runtime()()
    return rt.build(class_instance)


from . import sdk_pb2_grpc
from . import sdk_pb2
from truffle.runtime import HOST, RuntimeType
import os 

APP_SOCK = (
    os.getenv("TRUFFLE_APP_SOCKET")
    if os.getenv("TRUFFLE_APP_SOCKET") is not None
    else "unix:///tmp/truffle_app.sock"
)
SDK_SOCK = (
    os.getenv("TRUFFLE_SDK_SOCKET")
    if os.getenv("TRUFFLE_SDK_SOCKET") is not None
    else "unix:///tmp/truffle_sdk.sock"
)

SHARED_DIR = (
    os.getenv("TRUFFLE_SHARED_DIR")
    if os.getenv("TRUFFLE_SHARED_DIR") is not None
    else "/root/shared"  # container default 1.31.25
)

if HOST is RuntimeType.TRUFFLE:
    if not os.getenv("TRUFFLE_APP_SOCKET") or not os.getenv("TRUFFLE_SDK_SOCKET"):
        raise Exception("TRUFFLE_APP_SOCKET and TRUFFLE_SDK_SOCKET must be set when using Truffle SDK")
        



class TruffleFile:
    def __init__(self, path: str, name: str):
        self.path = path
        self.name = name


class TruffleImage:
    def __init__(self, path: str, name: str):
        self.path = path
        self.name = name




import ast
import sys
import json
import zipfile
import importlib
import sysconfig
from pathlib import Path

from truffle.cli.shared import *
from truffle.common import get_logger

logger = get_logger()


def _get_png_dimensions(file_path):
    with open(file_path, "rb") as f:
        signature = f.read(8)  
        if signature != b"\x89PNG\r\n\x1a\n":
            raise ValueError("Not a valid PNG file: " + file_path)
        f.read(4) 
        if f.read(4) != b"IHDR":
            raise ValueError("Invalid PNG structure: header not found in " + file_path)

        width = int.from_bytes(f.read(4), "big")
        height = int.from_bytes(f.read(4), "big")
        return width, height

def _check_file(p : Path):
    name = p.name
    if name == "manifest.json":
        manifest = json.loads(p.read_text())
        required_keys = [
            'name', 'description', 'app_bundle_id', 'manifest_version'
        ]
        for key in required_keys:
            if key not in manifest:
                logger.error(f"Missing key {key} in manifest.json")
                sys.exit(1)
        # if 'developer_id' not in manifest: manifest['developer_id'] = get_user_id()
        if 'example_prompts' not in manifest or len(manifest['example_prompts']) == 0:
            logger.warning("No example prompts found in manifest.json, auto-selection performance may be poor")
    elif name == "requirements.txt":
        reqs = p.read_text().strip().split("\n")
        if len(reqs) == 0:
            logger.error("Empty/unparseable requirements.txt file")
            sys.exit(1)
    elif name == "main.py":
        main_text = p.read_text()
        if main_text.find("import truffle") == -1:
            logger.error("Missing import truffle in main.py")
            sys.exit(1)
        
        if main_text.find("class") == -1:
            logger.error("Missing class definition in main.py")
            sys.exit(1)
        
    elif name == "icon.png":
        try:
            w, h = _get_png_dimensions(p)
            MIN_ICON_SIZE = 128
            if w < MIN_ICON_SIZE or h < MIN_ICON_SIZE:
                logger.error(f"Invalid icon.png size: {w}x{h} - must be more than {MIN_ICON_SIZE}x{MIN_ICON_SIZE}")
                sys.exit(1)
        except Exception as e:
            logger.error(f"Invalid icon.png file: {e}")
            sys.exit(1)
    else:
        logger.error(f"Unknown file: {p}")
        sys.exit(1)
    logger.debug(f"Checked {p} - OK")

def _must_exist(p : Path):
    if not p.exists() or not p.is_file():
        logger.error(f"Missing file: {p} - invalid project")
        sys.exit(1)
    if p.stat().st_size == 0:
        logger.error(f"Empty file: {p} - invalid project")
        if p.name == "requirements.txt":
            logger.warning("if a requirements.txt file is not needed, please remove it")
        sys.exit(1)
    _check_file(p)

def _check_syntax_and_decorators(path: Path):
    src = path.read_text()
    # check syntax
    try: tree = ast.parse(src, filename=str(path))
    except SyntaxError as e:
        logger.error(f"SyntaxError in {e.filename}, line {e.lineno}: {e.msg}")
        return False

    for node in ast.walk(tree):
        if isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef, ast.ClassDef)):
            for deco in node.decorator_list:
                lineno = getattr(deco, "lineno", "?")

                # pull out the actual decorator name whether it's @mod.attr or @mod.attr(...)
                # will add checks for arguments later
                if isinstance(deco, ast.Call): func = deco.func
                else: func = deco

                # func should be an ast.Attribute if it's mod.attr
                if not isinstance(func, ast.Attribute): continue

                # ensure the “mod” side is a simple Name (e.g. “truffle”)
                if not isinstance(func.value, ast.Name): continue

                mod_name  = func.value.id
                attr_name = func.attr

                # try importing and hasattr
                try: module = importlib.import_module(mod_name)
                except ImportError:
                    logger.error(
                        f"ImportError: module '{mod_name}' not found "
                        f"(in decorator on '{node.name}', line {lineno})"
                    )
                    return False
                else:
                    if not hasattr(module, attr_name):
                        logger.error(
                            f"AttributeError: module '{mod_name}' has no "
                            f"attribute '{attr_name}' (in decorator on '{node.name}', line {lineno})"
                        )
                        return False            
    return True

def _load_requirements_pkgs(req_path: Path):
    pkgs = set()
    for line in req_path.read_text().splitlines():
        line = line.strip()
        if not line or line.startswith("#"): continue
        # strip off version specifiers
        name = line.split(";", 1)[0] # drop env markers
        name = name.split("==", 1)[0]
        name = name.split(">=", 1)[0]
        name = name.split("<=", 1)[0]
        name = name.split("!=", 1)[0]
        name = name.split("~=", 1)[0]
        name = name.split(">", 1)[0]
        name = name.split("<", 1)[0]
        pkgs.add(name.strip().lower())
    return pkgs

def _check_imports(path: Path, reqs_path: Path):
    src = path.read_text()
    try: tree = ast.parse(src, filename=str(path))
    except SyntaxError as e: 
        logger.error(f"SyntaxError when parsing imports: {e.msg} (line {e.lineno})")
        return False
    required_pkgs = _load_requirements_pkgs(reqs_path)
    stdlib_dir = sysconfig.get_paths()["stdlib"]
    seen = set()

    for node in ast.walk(tree):
        if isinstance(node, ast.Import):
            for alias in node.names: seen.add(alias.name.split(".", 1)[0])
        elif isinstance(node, ast.ImportFrom):
            if node.level == 0 and node.module: seen.add(node.module.split(".", 1)[0])

    for mod in sorted(seen):
        spec = importlib.util.find_spec(mod)
        if spec is None and mod not in BANNED_REQS:
            logger.error(f"Module not found at all: '{mod}'")
            return False

        origin = spec.origin or ""
        if origin == "built-in": continue
        if origin.startswith(stdlib_dir): continue
        # third-party → must be in requirements.txt
        if mod.lower() not in required_pkgs and mod.lower() not in BANNED_REQS:
            logger.error(f"Third-party module '{mod}' not listed in requirements.txt")
            return False
        logger.debug(f"-> found requirement '{mod}' - OK")

    for r in required_pkgs:
        if r.lower() in BANNED_REQS:
            logger.error(f"Do not include '{r}' in requirements.txt")
            return False

    return True


def _looks_like_venv(builddir: Path, dirname: str) -> bool:
    venv_path = builddir / dirname
    if not venv_path.is_dir(): return False
    markers = [
        venv_path / 'pyvenv.cfg',
        venv_path / 'bin' / 'activate',
        venv_path / 'bin' / 'activate.fish',
        venv_path / 'bin' / 'activate.csh',
        venv_path / 'Scripts' / 'activate.bat'
    ]
    return any(marker.exists() for marker in markers)

def _make_zip(builddir: Path, dst_file: Path):
    blacklist_files = ['.DS_Store', '.gitignore']
    blacklist_dirs = ['__pycache__', '.git']
    assert builddir.exists() and builddir.is_dir(), f"Invalid source directory: {builddir}"
    assert dst_file.suffix == ".truffle", f"Invalid destination file: {dst_file}"
    with zipfile.ZipFile(dst_file, 'w', zipfile.ZIP_DEFLATED) as zipf:
        for root, dirs, files in os.walk(builddir):
            original_dirs = list(dirs)
            dirs[:] = [d for d in dirs if d not in blacklist_dirs and not _looks_like_venv(builddir, d)]
            for d in original_dirs:
                if d not in dirs:
                    logger.warning(f"Skipping directory '{d}', matches blacklist or venv pattern")
            zipf.write(root, arcname=os.path.join(builddir.name, os.path.relpath(root, builddir)))
            for file in files:
                if file.endswith('.truffle'): continue
                if file in blacklist_files: continue
                file_path = os.path.join(root, file)
                arcname = os.path.join(builddir.name, os.path.relpath(file_path, builddir))
                zipf.write(file_path, arcname)
                logger.debug(f"Added file to zip: {arcname}")
    bundle_size = dst_file.stat().st_size
    if bundle_size > 10 * 1024 * 1024:
        logger.error(f"Bundle size too large ({bundle_size/(1024 * 1024):.2f} MB > 10 MB), consider deleting some files")
        dst_file.unlink()
        return None
    return dst_file

def build(builddir: Path, from_upload: bool = False):
    if not builddir.exists():
        logger.error(f"Path {builddir} does not exist - cannot build")
        sys.exit(1)
    if not builddir.is_dir():
        logger.error("Path is not a directory, please provide a directory to build")
        sys.exit(1)

    for f in APP_FILES:
        _must_exist(builddir / f)

    for file in builddir.iterdir():
        try:
            if file.is_file() and not file.name.endswith(".truffle"):
                size = file.stat().st_size
                if size > (1024 * 1024 * 10): 
                    logger.error(f"Unexpectedly large file {file}, did you mean to include this?")
                    sys.exit(1)
        except Exception as e: continue
    
    main_path = builddir / "main.py"
    reqs_txt_path = builddir / "requirements.txt"

    if not os.getenv("TRUFFLE_SKIP_VALIDATE"):
        # check if truffle.run() exists
        inst_code = None
        for node in ast.walk(ast.parse(main_path.read_text())):
            if isinstance(node, ast.Call):
                func = node.func
                if (
                    isinstance(func, ast.Attribute) and func.attr == "run" and
                    isinstance(func.value, ast.Name) and func.value.id == "truffle"
                ):
                    inst_code = ast.get_source_segment(main_path.read_text(), node)
                    break
        if not inst_code:
            logger.error("Unable to find truffle.run(...) in main.py")
            sys.exit(1)

        # check for syntax and decorators
        logger.info("Checking main.py for syntax + decorators")
        if not _check_syntax_and_decorators(main_path):
            logger.error("Found errors, aborting.")
            exit(1)
        else: logger.success("No syntax or decorator errors detected.")

        # checking requirements
        logger.info("Checking requirements and imports")
        if not _check_imports(main_path, reqs_txt_path):
            logger.error("Found errors, aborting.")
            exit(1)
        else: logger.success("No errors in requiremenets.txt")

    bundle = _make_zip(builddir, builddir / f"{builddir.name}.truffle")
    if bundle == None: exit(1)
    try: rel_bundle = bundle.relative_to(Path.cwd())
    except ValueError: rel_bundle = bundle
    try:
        rel_dir = builddir.relative_to(Path.cwd())
        rel_upload_dir = f" {rel_dir}" if str(rel_dir) != "." else ""
    except ValueError: rel_upload_dir = f" {builddir}"
    logger.success(f"Built project {builddir.name} to {rel_bundle}")
    if not from_upload: 
        logger.info(f"Upload with 'truffle upload{rel_upload_dir}'")
        sys.exit(0)


import os
import sys
import webbrowser
from pathlib import Path

from truffle.common import get_logger

logger = get_logger()

APP_FILES = ["manifest.json", "requirements.txt", "main.py", "icon.png"]
BANNED_REQS = ["truffle", "grpcio", "protobuf"]

def get_client_userdata_dir() -> Path:
    if sys.platform == "win32": base = os.getenv("APPDATA", os.path.expanduser("~\\AppData\\Roaming"))
    elif sys.platform == "darwin": base = os.path.expanduser("~/Library/Application Support")
    else: base = os.getenv("XDG_CONFIG_HOME", os.path.expanduser("~/.config"))

    CLIENT_DIR_NAME = "TruffleOS-Development" if os.getenv("TRUFFLE_DEV") else "TruffleOS"
    userdata = Path(os.path.join(base, CLIENT_DIR_NAME))
    if not userdata.exists():
        logger.error(f"User data directory {userdata} does not exist")
        logger.error("Please make sure you have Truffle OS installed")
        logger.info("Opening download link... (https://itsalltruffles.com/)")
        webbrowser.open("https://itsalltruffles.com/", new=0, autoraise=True)
        exit(1)
    if not userdata.is_dir(): raise ValueError(f"User data directory {userdata} is not a directory")
    return userdata


def get_user_id() -> str:
    if os.getenv("TRUFFLE_CLIENT_ID"):
        return os.getenv("TRUFFLE_CLIENT_ID")
    try:
        userdata = get_client_userdata_dir()
        user_id_path = userdata / "magic-number.txt"
        if not user_id_path.exists():
            raise ValueError(f"Magic Number file @{user_id_path} does not exist")
        with open(user_id_path, 'r') as f:
            user_id = f.read().strip()
        if not user_id or len(user_id) < 6:
            raise ValueError(f"Magic Number file @{user_id_path} is empty/too short {len(user_id)}")
        
        if not user_id.isdigit():
            raise ValueError(f"Magic Number file @{user_id_path} is not a number")
        if user_id == "1234567891":
            raise ValueError(f"Magic Number file @{user_id_path} is the placeholder number")
        return user_id
    except Exception as e:
        logger.error(f"Error getting user ID: {e}")
        raise


def get_base_url() -> str:
    if os.getenv("TRUFFLE_DEV"): return "https://forecast.itsalltruffles.com:2087" 
    url = "https://overcast.itsalltruffles.com:2087"
    try:
       path = get_client_userdata_dir()
       url_file = path / "current-url"
       if url_file.exists():
            with open(url_file, "r") as f:
                url_file_contents = f.read().strip()
                if url_file_contents:
                    url = url_file_contents
                    logger.debug(f"Using URL from file: {url}")
    except Exception as e:
        logger.error(f"Error getting base URL fron client: {str(e)}")
    logger.debug(f"Using URL-  {url}")
    return url


import os
import sys
import json
import uuid
import base64
import zipfile
import requests
from  pathlib import Path

from truffle.cli.shared import *
from truffle.cli.build import build
from truffle.common import get_logger
from truffle.cli.color_argparse import ColorArgParser
from truffle.platform.sdk_pb2 import AppUploadProgress

# good imports are always 1/x shaped

logger = get_logger()


def default_mainpy(proj_name : str) -> str:
    return str(r"""import truffle
import requests
from typing import Dict

class PROJECT_NAME:
    def __init__(self):
        # The client for using the SDK API
        # This allows you to perform LLM inference, text embeddings, ask the user for input, etc.
        self.client = truffle.TruffleClient()   

        # You can store state in your class, it will be saved between tool calls and by the backend to reload saved tasks!            
        self.notepad = ""


    # The icon is from Apple SF Symbols, you can find a full list here: https://hotpot.ai/free-icons
    @truffle.tool(description="A Hello World tool", icon="smiley")
    @truffle.args(example_str="The text to print after Hello World")   # Add descriptions to your arguments to help the model
    def HelloWorld(self, example_str: str) -> Dict[str, str]:  # You have to type annotate all arguments and the return type
        return { "response" : "Hello World " + example_str  }

    
    # You can call external APIs easily
    @truffle.tool("Returns a joke", icon="face-smile")
    def GetJoke(self, num_jokes : int) -> str:
        num_jokes = 1 if num_jokes < 1 else num_jokes
        response = requests.get(f"https://v2.jokeapi.dev/joke/Programming,Misc?format=txt&amount={num_jokes}")
        if response.status_code != 200:
            # Any logs or exceptions you raise are forwaded to the model and the client
            print("JokeAPI returned an error: ", response.status_code)
            raise ValueError("JokeAPI is down, try again later")   
        return response.content
    

    # An example of tools using state! You might want to use this to store things like API keys, or user preferences
    @truffle.tool("Take notes", icon="pencil")
    def TakeNote(self, note: str) -> str:
        self.notepad += note + "\n"
        return "Added note.\n Current notes: \n" + str(self.notepad)


    # We will use this function as a predicate for the ReadNotes tool
    # Predicates should take no arguments and return a boolean
    def _notes_exist(self) -> bool:
        return self.notepad != ""
    

    # This tool will only be visible to the model if _NotesExist returns True
    @truffle.tool("Read notes", icon="glasses", predicate=_notes_exist)
    @truffle.args(clear_after="Whether to clear notes after reading.")
    def ReadNotes(self, clear_after: bool) -> str:
        notes = self.notepad
        if clear_after is True:
            self.notepad = ""
        return "Current notes: \n" + str(notes)
    

    # Here we create an async tool, so the model can do other things while the tool is running
    @truffle.tool("Searches with Perplexity", icon="magnifying-glass")
    @truffle.args(query="The search query")
    @truffle.flags(nonblock=True)   # This will make the tool async 
    def AsyncPerplexitySearch(self, query: str) -> str:    
        self.client.tool_update("Searching perplexity...")  # Send an update to the client, will be displayed in the UI
        return self.client.perplexity_search(query=query, model="sonar-pro")    # SDK API provides free access to Perplexity!
    
    
    # You can add as many tools as you want to your app, just make sure they are all in the same class, and have the @truffle.tool decorator
    # Of course, you may also add any other methods you want to your class, they will not be exposed to the model but can be used in your tools
    # Any files in your project directory will be included in the bundle, so you can use them in your tools as well, use relative paths from main.py
        
    # Have fun building!

if __name__ == "__main__":
    truffle.run(PROJECT_NAME())
""").replace("PROJECT_NAME", proj_name)


def upload(path: Path):
    if not path.exists():
        logger.error(f"Path {path} does not exist, cannot upload")
        sys.exit(1)
    
    if path.is_dir():
        truffle_files = [f for f in os.listdir(path) if f.endswith('.truffle')]
        if not truffle_files:
            logger.warning("No .truffle file found in the directory, attempting build...")
            build(path, from_upload=True)
            print()
            upload(path) # this exists with 0, but just in case
            sys.exit(0)
        if len(truffle_files) > 1:
            logger.error("Multiple .truffle files found in the directory")
            sys.exit(1)
        path = path / truffle_files[0]
    elif path.is_file():
        if not path.suffix == '.truffle':
            logger.error("File does not have a .truffle extension")
            sys.exit(1)
        if not zipfile.is_zipfile(path):
            logger.error("File is not a valid zip archive")
            sys.exit(1)
    else:
        logger.error("Path is neither a directory nor a file") # is this possible?
        sys.exit(1)
    
    logger.info(f"Uploading project: {path}")

    def post_zip_and_log_sse(url : str, zip_path : Path, user_id : str) -> bool:
        headers = {'user_id': user_id}
        assert zip_path.exists(), f"path {zip_path} does not exist"
        assert zip_path.is_file(), f"path {zip_path} is not a file"
        assert zip_path.suffix == '.truffle', f"zip_path {zip_path} is not a truffle file"
        with open(zip_path, 'rb') as f:
            files = {'file': ( zip_path.stem, f, 'application/zip')}
            with requests.post(url + "/install", headers=headers, files=files, stream=True, timeout=120) as resp:
                if resp.status_code != 200: raise requests.exceptions.HTTPError(f"Error: {resp.status_code} {resp.text}")
                buffer = bytes()
                for line_bytes in resp.iter_lines(decode_unicode=False, chunk_size=1024):
                    line = line_bytes.decode('utf-8', errors='replace')
                    logger.debug(f"decode line: ''{line}''")
                    if not line:
                        if buffer:
                            try: #deserialize the event
                                event_bytes = buffer
                                prog = AppUploadProgress()
                                decoded = base64.b64decode(event_bytes)
                                while decoded.endswith(b'\0'): decoded = decoded[:-1] # remove trailing null bytes 
                                prog.ParseFromString(decoded)
                                
                                if prog.substep: logger.info(f"{prog.substep}")
                                if str(prog.error) != "": logger.error(f"Error: '{prog.error}'")
                                if prog.step == AppUploadProgress.UploadStep.STEP_INSTALLED: return True
    
                            except Exception as e:
                                # long logs still might cause a parsing error, will be fixed later
                                # assume you can keep reading on past the occasional bad msg until the connection closes/ network errors happen
                                # FIXME, just hiding for now, will be fixed later from fw
                                # logger.error(f"Error parsing event pb: {e}")
                                logger.debug(f"Caught error: {e}")
                                logger.debug(f"buffer: <<<\n{buffer}\n>>>")
                                logger.debug(f"line: <<<\n{buffer.decode('utf-8',errors='replace')}\n>>>")
                            buffer = bytes()
                        continue
                    
                    # basic SSE format is data: <data> \n\n (so \n\n delimiter and data: prefix)
                    # this is not a great way to parse this but for sake of example - dont copy paste it (what if I did)
                    if line.startswith('data:'): buffer += line_bytes[5:].lstrip()
                    if 'error:' in line: logger.error(f"Error: {line}")
        return False

    URL = get_base_url()
    user_id = get_user_id()
    logger.debug(f"Uploading to {URL} with user ID {user_id}")

    try:
        if(post_zip_and_log_sse(URL, path, user_id)):
            logger.success("Upload successful, check your client for installation errors and confirmation")
            sys.exit(0)
        else:
            logger.error("Upload failed, please try again")
            sys.exit(1)
    except Exception as e:
        logger.error(f"Upload failed: unknown error: {e}")
        sys.exit(1)


def init_project(project : Path):
    if project.exists() and not project.is_dir():
            logger.error(f"Path {project} exists and is not a directory, cannot init")
            sys.exit(1)
    
    def alphanum_camel(string: str) -> str:
        return ''.join(w.capitalize() for w in string.split() if w.isalnum())

    def get_input(prompt : str, default : str = None, field : str = 'Input'):
        has_default = default is not None and len(default) > 0
        txt = logger.input(f"{prompt}: ") if not has_default else logger.input(f"{prompt} ({default}): ")
        if len(txt) > 0:
            return txt
        elif len(txt) == 0 and has_default:
            return default
        else:
            logger.error(f"{field} cannot be empty!")
            return get_input(prompt, default)
    
    logger.info("Provide some information about your project: ")
    proj_name = get_input(f"Name for your app", project.name, "App name")
    proj_desc = get_input("Description for your app", "Does app things", "App description")
    example_prompts = []
    logger.info("(optional) Provide some example prompts for your app: ")
    logger.info("These will be used for the automatic app selection feature, and can be edited in the manifest.json post-init")
    while True:
        msg = "Enter an example prompt for your app (or leave empty to finish): " if len(example_prompts) == 0 else "Enter another example prompt for your app (or leave empty to finish): "
        ex = logger.input(msg)
        if len(ex) == 0: break
        example_prompts.append(ex)
    
    print('')

    # If user runs in non empty dir, create a file with project name in CamelCase, increment number if dir exists
    # else if the dir is empty or doesn't exist then create the files there
    if project.exists() and any(project.iterdir()):
        for i in range(999):
            new_path = Path(project / (alphanum_camel(proj_name)+(str(i) if i>0 else '')))
            if not new_path.exists():
                project = new_path
                break
            i += 1
    project.mkdir(parents=True)

    for f in APP_FILES:
        if f == "icon.png":
            try:
                c = requests.get("https://pngimg.com/d/question_mark_PNG68.png")
                if c.status_code != 200: raise requests.HTTPError(c.status_code)
                with open(project / f, "wb") as icon: icon.write(c.content)
            except Exception as e:
                logger.error(f"Failed to download and create default icon: {e}")
                logger.warning("Creating a placeholder icon, please change it manually")
                (project / f).touch()
            logger.debug(f"-> Created default icon.png")
        elif f == "manifest.json":
            manifest = {
                "name": proj_name,
                "description": proj_desc,
                "example_prompts": example_prompts,
                "packages": [],
                "manifest_version": 1,
                "app_bundle_id" : str(uuid.uuid4()) # we can't just find your app by name, we need a unique identifier in case you change it! 
            }
            with open(project / f, "w") as mani:
                mani.write(json.dumps(manifest, indent=4, sort_keys=False, ensure_ascii=False))
            logger.debug(f"-> Created manifest.json with:\n {manifest}")
        elif f == "requirements.txt":
            with open(project / f, "w") as reqs: reqs.write("requests\n")
            logger.debug(f"-> Created requirements.txt")
        elif f == "main.py":
            class_name = alphanum_camel(proj_name)
            with open(project / f, "w") as main: main.write(default_mainpy(class_name))
            logger.debug(f"-> Created main.py with default content")
        else:
            (project / f).touch()

    logger.success(f"Initialized project in {project.resolve()}")
    try: rel_path = project.relative_to(Path.cwd())
    except ValueError: rel_path = project
    logger.info(f"You can now build your project with 'truffle build {rel_path}', and upload it with 'truffle upload {rel_path}'")
    sys.exit(0)


def cli():
    parser = ColorArgParser(prog="truffle", description="Truffle SDK CLI")
    subparsers = parser.add_subparsers(dest="action", description="the CLI action: upload / init / or build", required=True, help="The action to perform")
    
    def add_subcommand(name: str, help: str):
        p = subparsers.add_parser(name, help=help)
        p.add_argument("path", help="The path to the project", nargs="?", default=os.getcwd())
    
    actions = {
         "upload" : {"help": "Upload the project to the cloud", "fn": upload},
         "init" : {"help": "Initialize a new project", "fn": init_project},
         "build" : {"help": "Build the project", "fn": build},
    }

    for name, args in actions.items():
        add_subcommand(name, args["help"])

    args = parser.parse_args()
    if not args.action or args.action not in actions:
        parser.print_help()
        logger.error(f"\t-> please provide one of the following actions: {', '.join(actions.keys())}")
        sys.exit(1)
    
    cmd = actions[args.action]
    cmd["fn"](Path(args.path).resolve()) # resolve() fixes . as an arg
    

def main():
    cli()

if __name__ == "__main__":
    main()



from enum import Enum
import os

class RuntimeType(Enum):
    DEV = 0,
    CLIENT = 1,
    TRUFFLE = 2

def determine_runtime() -> RuntimeType:
    
    rt_envvar = os.getenv("TRUFFLE_RUNTIME")

    if rt_envvar == "CLIENT": return RuntimeType.CLIENT
    elif rt_envvar == "DEV": return RuntimeType.DEV
    elif rt_envvar == "TRUFFLE": return RuntimeType.TRUFFLE
    else: return RuntimeType.CLIENT




# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

from . import sdk_pb2 as sdk__pb2

GRPC_GENERATED_VERSION = '1.68.0'
GRPC_VERSION = grpc.__version__
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    raise RuntimeError(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + f' but the generated code in sdk_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
    )


class TruffleSDKStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Infer = channel.unary_stream(
                '/sdk.TruffleSDK/Infer',
                request_serializer=sdk__pb2.GenerateRequest.SerializeToString,
                response_deserializer=sdk__pb2.TokenResponse.FromString,
                _registered_method=True)
        self.InferSync = channel.unary_unary(
                '/sdk.TruffleSDK/InferSync',
                request_serializer=sdk__pb2.GenerateRequest.SerializeToString,
                response_deserializer=sdk__pb2.GenerateResponse.FromString,
                _registered_method=True)
        self.GetModels = channel.unary_unary(
                '/sdk.TruffleSDK/GetModels',
                request_serializer=sdk__pb2.GetModelsRequest.SerializeToString,
                response_deserializer=sdk__pb2.GetModelsResponse.FromString,
                _registered_method=True)
        self.Embed = channel.unary_unary(
                '/sdk.TruffleSDK/Embed',
                request_serializer=sdk__pb2.EmbedRequest.SerializeToString,
                response_deserializer=sdk__pb2.EmbedResponse.FromString,
                _registered_method=True)
        self.AskUser = channel.unary_unary(
                '/sdk.TruffleSDK/AskUser',
                request_serializer=sdk__pb2.UserRequest.SerializeToString,
                response_deserializer=sdk__pb2.UserResponse.FromString,
                _registered_method=True)
        self.ToolUpdate = channel.unary_unary(
                '/sdk.TruffleSDK/ToolUpdate',
                request_serializer=sdk__pb2.ToolUpdateRequest.SerializeToString,
                response_deserializer=sdk__pb2.SDKResponse.FromString,
                _registered_method=True)
        self.SystemTool = channel.unary_unary(
                '/sdk.TruffleSDK/SystemTool',
                request_serializer=sdk__pb2.SystemToolRequest.SerializeToString,
                response_deserializer=sdk__pb2.SystemToolResponse.FromString,
                _registered_method=True)
        self.GlobalVars = channel.unary_unary(
                '/sdk.TruffleSDK/GlobalVars',
                request_serializer=sdk__pb2.GlobalsRequest.SerializeToString,
                response_deserializer=sdk__pb2.GlobalsResponse.FromString,
                _registered_method=True)


class TruffleSDKServicer(object):
    """Missing associated documentation comment in .proto file."""

    def Infer(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def InferSync(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetModels(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Embed(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def AskUser(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ToolUpdate(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SystemTool(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GlobalVars(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_TruffleSDKServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Infer': grpc.unary_stream_rpc_method_handler(
                    servicer.Infer,
                    request_deserializer=sdk__pb2.GenerateRequest.FromString,
                    response_serializer=sdk__pb2.TokenResponse.SerializeToString,
            ),
            'InferSync': grpc.unary_unary_rpc_method_handler(
                    servicer.InferSync,
                    request_deserializer=sdk__pb2.GenerateRequest.FromString,
                    response_serializer=sdk__pb2.GenerateResponse.SerializeToString,
            ),
            'GetModels': grpc.unary_unary_rpc_method_handler(
                    servicer.GetModels,
                    request_deserializer=sdk__pb2.GetModelsRequest.FromString,
                    response_serializer=sdk__pb2.GetModelsResponse.SerializeToString,
            ),
            'Embed': grpc.unary_unary_rpc_method_handler(
                    servicer.Embed,
                    request_deserializer=sdk__pb2.EmbedRequest.FromString,
                    response_serializer=sdk__pb2.EmbedResponse.SerializeToString,
            ),
            'AskUser': grpc.unary_unary_rpc_method_handler(
                    servicer.AskUser,
                    request_deserializer=sdk__pb2.UserRequest.FromString,
                    response_serializer=sdk__pb2.UserResponse.SerializeToString,
            ),
            'ToolUpdate': grpc.unary_unary_rpc_method_handler(
                    servicer.ToolUpdate,
                    request_deserializer=sdk__pb2.ToolUpdateRequest.FromString,
                    response_serializer=sdk__pb2.SDKResponse.SerializeToString,
            ),
            'SystemTool': grpc.unary_unary_rpc_method_handler(
                    servicer.SystemTool,
                    request_deserializer=sdk__pb2.SystemToolRequest.FromString,
                    response_serializer=sdk__pb2.SystemToolResponse.SerializeToString,
            ),
            'GlobalVars': grpc.unary_unary_rpc_method_handler(
                    servicer.GlobalVars,
                    request_deserializer=sdk__pb2.GlobalsRequest.FromString,
                    response_serializer=sdk__pb2.GlobalsResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'sdk.TruffleSDK', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('sdk.TruffleSDK', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class TruffleSDK(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def Infer(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(
            request,
            target,
            '/sdk.TruffleSDK/Infer',
            sdk__pb2.GenerateRequest.SerializeToString,
            sdk__pb2.TokenResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def InferSync(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/sdk.TruffleSDK/InferSync',
            sdk__pb2.GenerateRequest.SerializeToString,
            sdk__pb2.GenerateResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetModels(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/sdk.TruffleSDK/GetModels',
            sdk__pb2.GetModelsRequest.SerializeToString,
            sdk__pb2.GetModelsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Embed(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/sdk.TruffleSDK/Embed',
            sdk__pb2.EmbedRequest.SerializeToString,
            sdk__pb2.EmbedResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def AskUser(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/sdk.TruffleSDK/AskUser',
            sdk__pb2.UserRequest.SerializeToString,
            sdk__pb2.UserResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ToolUpdate(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/sdk.TruffleSDK/ToolUpdate',
            sdk__pb2.ToolUpdateRequest.SerializeToString,
            sdk__pb2.SDKResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def SystemTool(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/sdk.TruffleSDK/SystemTool',
            sdk__pb2.SystemToolRequest.SerializeToString,
            sdk__pb2.SystemToolResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GlobalVars(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/sdk.TruffleSDK/GlobalVars',
            sdk__pb2.GlobalsRequest.SerializeToString,
            sdk__pb2.GlobalsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



from . import platform
from .platform import SDK_SOCK
from .common import get_logger


import grpc
import json
import typing


logger = get_logger()

#02.01.25
#this sucks!!! derek fucked it
# should just be methods, and inference needs really good abstraction
#client can be a fucking global set by runtime
# fix this, fix trufflefile, and add client runtime that validates!

#then CLI and this shit is good to go


class TruffleClient:
    def __init__(self, host=SDK_SOCK):
        self.channel = grpc.insecure_channel(host)
        self.stub = platform.sdk_pb2_grpc.TruffleSDKStub(self.channel)
        self.model_contexts: list[platform.sdk_pb2.Context] = []

    def perplexity_search(
        self,
        query: str,
        model: str = "sonar",
        response_fmt=None,
        system_prompt: str = "",
    ) -> str:
        # https://docs.perplexity.ai/guides/model-cards
        perplexity_models_feb24 = [
            "sonar-reasoning",  # Chat Completion - 127k context length
            "sonar-pro",        # Chat Completion - 200k 
            "sonar",            # Chat Completion - 127k 
        ]
        if model not in perplexity_models_feb24:
            raise ValueError(
                f"Model '{model}' not found in available models [{perplexity_models_feb24}]. See https://docs.perplexity.ai/guides/model-cards"
            )

        PERPLEXITY_API_URL = "https://api.perplexity.ai/chat/completions"
        payload = {
            "model": model,
            "messages": [
                {"role": "system", "content": system_prompt},
                {"role": "user", "content": query},
            ],
        }
        # https://docs.perplexity.ai/guides/structured-outputs
        if response_fmt is not None:
            payload["response_format"] = response_fmt

        # TODO: support everything! they have such a cool API - image return would be insane, and citations!
        # https://docs.perplexity.ai/api-reference/chat-completions
        try:
            request = platform.sdk_pb2.SystemToolRequest(tool_name="perplexity_search")
            request.args["url"] = PERPLEXITY_API_URL
            request.args["payload"] = json.dumps(payload)

            r: platform.sdk_pb2.SystemToolResponse = self.stub.SystemTool(request)
            if r.error:
                raise RuntimeError(f"SystemToolError: {r.error}")

            results = json.loads(r.response)

            return results["choices"][0]["message"]["content"]
        except grpc.RpcError as e:
            raise RuntimeError(f"RPC error: {e.details()}")
        except json.JSONDecodeError as e:
            raise RuntimeError(f"JSON error: {e}")

    def get_models(self):
        response: list[platform.sdk_pb2.ModelDescription] = self.stub.GetModels(
            platform.sdk_pb2.GetModelsRequest()
        ).models
        return response

    # WIP - missing ui
    def tool_update(self, message: str):
        try:
            r: platform.sdk_pb2.SDKResponse = self.stub.ToolUpdate(
                platform.sdk_pb2.ToolUpdateRequest(friendly_description=message)
            )
            if r.error:
                raise RuntimeError(f"RPC error: {r.error}")

        except grpc.RpcError as e:
            raise RuntimeError(f"RPC error: {e.details()}")

    # WIP - missing ui
    def ask_user(
        self, message: str, reason: str = "Tool needs input to continue."
    ) -> typing.Dict[str, typing.Union[str, typing.List[str]]]:
        
        try:
            response: platform.sdk_pb2.UserResponse = self.stub.AskUser(
                platform.sdk_pb2.UserRequest(message=message, reason=reason)
            )
            ret = {"response": response.response}
            return ret

        except grpc.RpcError as e:
            raise RuntimeError(f"RPC error: {e.details()}")

    def query_embed(
        self, query: str, documents: typing.List[str]
    ) -> typing.List[typing.Tuple[str, float]]:
        """
        Returns a list of documents sorted by cosine similarity to the query via. text embeddings, which really should include that value. Doh.

        Args:
            query: The query string
            documents: A list of document strings to search

        Returns:
            A list of document strings sorted by cosine similarity to the query from most to least similar

        """
        request = platform.sdk_pb2.EmbedRequest(query=query, documents=documents)
        # i have not tested this
        try:
            response: platform.sdk_pb2.EmbedResponse = self.stub.Embed(request)
            print("Embedding response: ")
            results = []
            if len(response.results) == 0:
                raise ValueError("No results returned")
            for r in response.results:
                print(f"{r.text}: {r.score}")
                results.append((r.text, r.score))
            return results
        except grpc.RpcError as e:
            raise RuntimeError(f"Embed RPC error: {e.details()}")
        except ValueError as e:
            raise RuntimeError(f"Embed Value error: {e}")

    def infer(
        self,
        prompt: str,
        model_id: int = 0,
        system_prompt: str | None = None,
        context_idx: int | None = None,
        max_tokens: int = 1000,
        temperature: float = 0.7,
        format_type: str = None,
        schema: str = None,
    ) -> typing.Iterator[str]:
        """
        Make a streaming inference request to the TruffleSDK service.

        Args:
            prompt: The input prompt for generation
            max_tokens: Maximum number of tokens to generate
            temperature: Sampling temperature (0.0 to 1.0)
            format_type: Optional response format (TEXT, JSON, EBNF)
            schema: Optional schema for structured output

        Returns:
            Iterator yielding generated tokens
        """
        # TODO: properly handle format stuff
        format_spec = None
        if format_type:
            format_spec = platform.sdk_pb2.GenerateResponseFormat(
                format=platform.sdk_pb2.GenerateResponseFormat.ResponseFormat.Value(
                    f"RESPONSE_{format_type}"
                ),
                schema=schema,
            )

        # Fetch or build the context for the chat
        if context_idx is not None:
            if self.context_idx is not None and system_prompt is not None:
                raise ValueError(
                    "Only pass system_prompt or context_idx, but not both!"
                )
            current_context = self.model_contexts[context_idx]
        else:
            self.model_contexts.append(platform.sdk_pb2.Context())
            current_context: platform.sdk_pb2.Context = self.model_contexts[-1]
            if system_prompt is not None:
                current_context.history.append(
                    platform.sdk_pb2.Content(
                        role=platform.sdk_pb2.Content.ROLE_SYSTEM, content=system_prompt
                    )
                )

        current_context.history.append(
            platform.sdk_pb2.Content(
                role=platform.sdk_pb2.Content.Role.ROLE_USER, content=prompt
            )
        )

        # Create the generation request
        request = platform.sdk_pb2.GenerateRequest(
            model_id=model_id,
            context=current_context,
            max_tokens=max_tokens,
            temperature=temperature,
            fmt=format_spec,
        )

        try:
            # Make the streaming call
            streamed_message = []
            for response in self.stub.Infer(request):
                if response.error:
                    raise RuntimeError(f"Generation error: {response.error}")

                if response.finish_reason:
                    # Handle different finish reasons
                    if (
                        response.finish_reason
                        == platform.sdk_pb2.GenerateFinishReason.FINISH_REASON_ERROR
                    ):
                        raise RuntimeError("Generation terminated with error")
                    elif (
                        response.finish_reason
                        != platform.sdk_pb2.GenerateFinishReason.FINISH_REASON_UNSPECIFIED
                    ):
                        break

                streamed_message.append(response.token)
                yield response.token

            current_context.history.append(
                platform.sdk_pb2.Content(
                    role=platform.sdk_pb2.Content.ROLE_AI,
                    content="".join(streamed_message),
                )
            )

        except grpc.RpcError as e:
            raise RuntimeError(f"RPC error: {e} {e.details()}")


    def _globalvar(self, key:str, value:str = None) -> str:
        if value is not None:
            value = str(value)
        request = platform.sdk_pb2.GlobalsRequest(key=key, value=value)
        logger.debug(f"GlobalsRequest: {request}")
        response: platform.sdk_pb2.GlobalsResponse = self.stub.GlobalVars(request)
        logger.debug(f"GlobalsResponse: {response}")
        return response.value
    def close(self):
        """Close the gRPC channel"""
        self.channel.close()


class TGlobals:
    _reserved = ['set', 'get', 'contains', '_client', '_cache']
    def __init__(self, client: TruffleClient = TruffleClient()):
        self._client = client
        self._cache : typing.Dict[str, bool] = {}
    def __getattr__(self, key: str) -> str: #can throw 
       return self.get(key)
    def __setattr__(self, key: str, value: str) -> None:
        if key in self._reserved:
            super().__setattr__(key, value)   # normal attribute write
            return
        self.set(key, value)
       
    
    def __getitem__(self, key):
        return self.__getattr__(key)
    def __setitem__(self, key, value):
        return self.__setattr__(key, value)
    def __contains__(self, key: str) -> bool:
        return self.contains(key)
    
    def set(self, key: str, value: str) -> str:
        #returns prev value if set 
        prev = self._client._globalvar(key, value)
        self._cache[key] = True
        return prev
    def contains(self, key: str) -> bool:
        if key in self._cache:
            return True
        try:
            v = self.get(key)
            if v is not None:
                self._cache[key] = True
                return True
        except Exception as e:
            pass
        return False
     
    
    def get(self, key: str) -> str:
        #returns prev value if set 
        return self._client._globalvar(key)
    




# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: sdk.proto
# Protobuf Python Version: 5.28.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    28,
    1,
    '',
    'sdk.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\tsdk.proto\x12\x03sdk\"\xa9\x03\n\x10ModelDescription\x12\x10\n\x08model_id\x18\x01 \x01(\x05\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x03 \x01(\t\x12\x38\n\x0c\x63\x61pabilities\x18\x04 \x01(\x0b\x32\".sdk.ModelDescription.Capabilities\x12-\n\x04type\x18\x05 \x01(\x0e\x32\x1f.sdk.ModelDescription.ModelType\x12\x10\n\x08is_local\x18\x06 \x01(\x08\x1a\x7f\n\x0c\x43\x61pabilities\x12\x19\n\x11structured_output\x18\x01 \x01(\x08\x12\x15\n\rimage_support\x18\x02 \x01(\x08\x12\x10\n\x08reasoner\x18\x03 \x01(\x08\x12\x18\n\x10world_domination\x18\x04 \x01(\x08\x12\x11\n\tlanguages\x18\x05 \x03(\t\"d\n\tModelType\x12\x15\n\x11MODEL_UNSPECIFIED\x10\x00\x12\x0f\n\x0bMODEL_SMART\x10\x01\x12\x0e\n\nMODEL_FAST\x10\x02\x12\x10\n\x0cMODEL_VISION\x10\x03\x12\r\n\tMODEL_AGI\x10\x04\"\x12\n\x10GetModelsRequest\":\n\x11GetModelsResponse\x12%\n\x06models\x18\x01 \x03(\x0b\x32\x15.sdk.ModelDescription\";\n\x0eGlobalsRequest\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x12\n\x05value\x18\x02 \x01(\tH\x00\x88\x01\x01\x42\x08\n\x06_value\" \n\x0fGlobalsResponse\x12\r\n\x05value\x18\x01 \x01(\t\"\x83\x01\n\x11SystemToolRequest\x12\x11\n\ttool_name\x18\x01 \x01(\t\x12.\n\x04\x61rgs\x18\x02 \x03(\x0b\x32 .sdk.SystemToolRequest.ArgsEntry\x1a+\n\tArgsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\xab\x01\n\x12SystemToolResponse\x12\x10\n\x08response\x18\x01 \x01(\t\x12\x35\n\x07results\x18\x02 \x03(\x0b\x32$.sdk.SystemToolResponse.ResultsEntry\x12\x12\n\x05\x65rror\x18\x03 \x01(\tH\x00\x88\x01\x01\x1a.\n\x0cResultsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x42\x08\n\x06_error\"7\n\x0bSDKResponse\x12\n\n\x02ok\x18\x01 \x01(\x08\x12\x12\n\x05\x65rror\x18\x02 \x01(\tH\x00\x88\x01\x01\x42\x08\n\x06_error\"1\n\x11ToolUpdateRequest\x12\x1c\n\x14\x66riendly_description\x18\x01 \x01(\t\".\n\x0bUserRequest\x12\x0f\n\x07message\x18\x01 \x01(\t\x12\x0e\n\x06reason\x18\x02 \x01(\t\"V\n\x0cUserResponse\x12\x10\n\x08response\x18\x01 \x01(\t\x12\x12\n\x05\x65rror\x18\x02 \x01(\tH\x00\x88\x01\x01\x12\x16\n\x0e\x61ttached_files\x18\x03 \x03(\tB\x08\n\x06_error\"\xac\x01\n\x16GenerateResponseFormat\x12:\n\x06\x66ormat\x18\x01 \x01(\x0e\x32*.sdk.GenerateResponseFormat.ResponseFormat\x12\x13\n\x06schema\x18\x02 \x01(\tH\x00\x88\x01\x01\"6\n\x0eResponseFormat\x12\x11\n\rRESPONSE_TEXT\x10\x00\x12\x11\n\rRESPONSE_JSON\x10\x01\x42\t\n\x07_schema\"\x9e\x01\n\x07\x43ontent\x12\x1f\n\x04role\x18\x01 \x01(\x0e\x32\x11.sdk.Content.Role\x12\x0f\n\x07\x63ontent\x18\x02 \x01(\t\x12\x11\n\x04\x64\x61ta\x18\x03 \x01(\x0cH\x00\x88\x01\x01\"E\n\x04Role\x12\x10\n\x0cROLE_INVALID\x10\x00\x12\r\n\tROLE_USER\x10\x01\x12\x0b\n\x07ROLE_AI\x10\x02\x12\x0f\n\x0bROLE_SYSTEM\x10\x03\x42\x07\n\x05_data\"(\n\x07\x43ontext\x12\x1d\n\x07history\x18\x01 \x03(\x0b\x32\x0c.sdk.Content\"\xd7\x01\n\x0fGenerateRequest\x12\x11\n\x07message\x18\x01 \x01(\tH\x00\x12\x1f\n\x07\x63ontext\x18\x02 \x01(\x0b\x32\x0c.sdk.ContextH\x00\x12\x12\n\nmax_tokens\x18\x03 \x01(\x05\x12\x10\n\x08model_id\x18\x04 \x01(\x05\x12\x18\n\x0btemperature\x18\x05 \x01(\x02H\x01\x88\x01\x01\x12-\n\x03\x66mt\x18\x06 \x01(\x0b\x32\x1b.sdk.GenerateResponseFormatH\x02\x88\x01\x01\x42\t\n\x07\x63ontentB\x0e\n\x0c_temperatureB\x06\n\x04_fmt\"X\n\x0fGenerationUsage\x12\x15\n\rprompt_tokens\x18\x01 \x01(\x05\x12\x19\n\x11\x63ompletion_tokens\x18\x02 \x01(\x05\x12\x13\n\x0b\x61pprox_time\x18\x03 \x01(\x05\"\xb9\x01\n\rTokenResponse\x12\r\n\x05token\x18\x01 \x01(\t\x12\x35\n\rfinish_reason\x18\x02 \x01(\x0e\x32\x19.sdk.GenerateFinishReasonH\x00\x88\x01\x01\x12(\n\x05usage\x18\x03 \x01(\x0b\x32\x14.sdk.GenerationUsageH\x01\x88\x01\x01\x12\x12\n\x05\x65rror\x18\x04 \x01(\tH\x02\x88\x01\x01\x42\x10\n\x0e_finish_reasonB\x08\n\x06_usageB\x08\n\x06_error\"\xbf\x01\n\x10GenerateResponse\x12\x10\n\x08response\x18\x01 \x01(\t\x12\x35\n\rfinish_reason\x18\x02 \x01(\x0e\x32\x19.sdk.GenerateFinishReasonH\x00\x88\x01\x01\x12(\n\x05usage\x18\x03 \x01(\x0b\x32\x14.sdk.GenerationUsageH\x01\x88\x01\x01\x12\x12\n\x05\x65rror\x18\x04 \x01(\tH\x02\x88\x01\x01\x42\x10\n\x0e_finish_reasonB\x08\n\x06_usageB\x08\n\x06_error\".\n\x0fSortedEmbedding\x12\x0c\n\x04text\x18\x01 \x01(\t\x12\r\n\x05score\x18\x02 \x01(\x02\"6\n\rEmbedResponse\x12%\n\x07results\x18\x01 \x03(\x0b\x32\x14.sdk.SortedEmbedding\"0\n\x0c\x45mbedRequest\x12\x11\n\tdocuments\x18\x01 \x03(\t\x12\r\n\x05query\x18\x02 \x01(\t\"\xc8\x05\n\x11\x41ppUploadProgress\x12/\n\x04step\x18\x01 \x01(\x0e\x32!.sdk.AppUploadProgress.UploadStep\x12\x10\n\x08progress\x18\x02 \x01(\x05\x12/\n\x04type\x18\x03 \x01(\x0e\x32!.sdk.AppUploadProgress.UploadType\x12\x13\n\x0blatest_logs\x18\x04 \x01(\t\x12\x14\n\x07substep\x18\x05 \x01(\tH\x00\x88\x01\x01\x12\x39\n\x05\x65rror\x18\n \x01(\x0b\x32%.sdk.AppUploadProgress.AppUploadErrorH\x01\x88\x01\x01\x1a#\n\x0e\x41ppUploadError\x12\x11\n\traw_error\x18\x01 \x01(\t\"\xba\x02\n\nUploadStep\x12\x10\n\x0cSTEP_UNKNOWN\x10\x00\x12\x16\n\x12STEP_UPLOAD_BUNDLE\x10\x01\x12\x16\n\x12STEP_VERIFY_BUNDLE\x10\x02\x12\x19\n\x15STEP_SYNC_WITH_SERVER\x10\x03\x12\x15\n\x11STEP_WRITE_BUNDLE\x10\x04\x12\x19\n\x15STEP_CREATE_CONTAINER\x10\x05\x12\x1d\n\x19STEP_INSTALL_DEPENDENCIES\x10\x06\x12!\n\x1dSTEP_GENERATE_CLASSIFIER_DATA\x10\x07\x12\x17\n\x13STEP_SAVE_CONTAINER\x10\x08\x12\x15\n\x11STEP_GATHER_TOOLS\x10\t\x12\x17\n\x13STEP_VERIFY_INSTALL\x10\n\x12\x12\n\x0eSTEP_INSTALLED\x10\x0b\"[\n\nUploadType\x12\x1c\n\x18UPLOAD_TYPE_UNDETERMINED\x10\x00\x12\x17\n\x13UPLOAD_TYPE_INSTALL\x10\x01\x12\x16\n\x12UPLOAD_TYPE_UPDATE\x10\x02\x42\n\n\x08_substepB\x08\n\x06_errorJ\x04\x08\x06\x10\n*\x98\x01\n\x14GenerateFinishReason\x12\x1d\n\x19\x46INISH_REASON_UNSPECIFIED\x10\x00\x12\x18\n\x14\x46INISH_REASON_LENGTH\x10\x01\x12\x16\n\x12\x46INISH_REASON_STOP\x10\x02\x12\x17\n\x13\x46INISH_REASON_ERROR\x10\x03\x12\x16\n\x12\x46INISH_REASON_USER\x10\x04*K\n\x0bTruffleType\x12\x17\n\x13TRUFFLE_UNSPECIFIED\x10\x00\x12\x11\n\rTRUFFLE_IMAGE\x10\x01\x12\x10\n\x0cTRUFFLE_FILE\x10\x02\x32\xc7\x03\n\nTruffleSDK\x12\x33\n\x05Infer\x12\x14.sdk.GenerateRequest\x1a\x12.sdk.TokenResponse0\x01\x12\x38\n\tInferSync\x12\x14.sdk.GenerateRequest\x1a\x15.sdk.GenerateResponse\x12:\n\tGetModels\x12\x15.sdk.GetModelsRequest\x1a\x16.sdk.GetModelsResponse\x12.\n\x05\x45mbed\x12\x11.sdk.EmbedRequest\x1a\x12.sdk.EmbedResponse\x12.\n\x07\x41skUser\x12\x10.sdk.UserRequest\x1a\x11.sdk.UserResponse\x12\x36\n\nToolUpdate\x12\x16.sdk.ToolUpdateRequest\x1a\x10.sdk.SDKResponse\x12=\n\nSystemTool\x12\x16.sdk.SystemToolRequest\x1a\x17.sdk.SystemToolResponse\x12\x37\n\nGlobalVars\x12\x13.sdk.GlobalsRequest\x1a\x14.sdk.GlobalsResponseb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'sdk_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_SYSTEMTOOLREQUEST_ARGSENTRY']._loaded_options = None
  _globals['_SYSTEMTOOLREQUEST_ARGSENTRY']._serialized_options = b'8\001'
  _globals['_SYSTEMTOOLRESPONSE_RESULTSENTRY']._loaded_options = None
  _globals['_SYSTEMTOOLRESPONSE_RESULTSENTRY']._serialized_options = b'8\001'
  _globals['_GENERATEFINISHREASON']._serialized_start=3111
  _globals['_GENERATEFINISHREASON']._serialized_end=3263
  _globals['_TRUFFLETYPE']._serialized_start=3265
  _globals['_TRUFFLETYPE']._serialized_end=3340
  _globals['_MODELDESCRIPTION']._serialized_start=19
  _globals['_MODELDESCRIPTION']._serialized_end=444
  _globals['_MODELDESCRIPTION_CAPABILITIES']._serialized_start=215
  _globals['_MODELDESCRIPTION_CAPABILITIES']._serialized_end=342
  _globals['_MODELDESCRIPTION_MODELTYPE']._serialized_start=344
  _globals['_MODELDESCRIPTION_MODELTYPE']._serialized_end=444
  _globals['_GETMODELSREQUEST']._serialized_start=446
  _globals['_GETMODELSREQUEST']._serialized_end=464
  _globals['_GETMODELSRESPONSE']._serialized_start=466
  _globals['_GETMODELSRESPONSE']._serialized_end=524
  _globals['_GLOBALSREQUEST']._serialized_start=526
  _globals['_GLOBALSREQUEST']._serialized_end=585
  _globals['_GLOBALSRESPONSE']._serialized_start=587
  _globals['_GLOBALSRESPONSE']._serialized_end=619
  _globals['_SYSTEMTOOLREQUEST']._serialized_start=622
  _globals['_SYSTEMTOOLREQUEST']._serialized_end=753
  _globals['_SYSTEMTOOLREQUEST_ARGSENTRY']._serialized_start=710
  _globals['_SYSTEMTOOLREQUEST_ARGSENTRY']._serialized_end=753
  _globals['_SYSTEMTOOLRESPONSE']._serialized_start=756
  _globals['_SYSTEMTOOLRESPONSE']._serialized_end=927
  _globals['_SYSTEMTOOLRESPONSE_RESULTSENTRY']._serialized_start=871
  _globals['_SYSTEMTOOLRESPONSE_RESULTSENTRY']._serialized_end=917
  _globals['_SDKRESPONSE']._serialized_start=929
  _globals['_SDKRESPONSE']._serialized_end=984
  _globals['_TOOLUPDATEREQUEST']._serialized_start=986
  _globals['_TOOLUPDATEREQUEST']._serialized_end=1035
  _globals['_USERREQUEST']._serialized_start=1037
  _globals['_USERREQUEST']._serialized_end=1083
  _globals['_USERRESPONSE']._serialized_start=1085
  _globals['_USERRESPONSE']._serialized_end=1171
  _globals['_GENERATERESPONSEFORMAT']._serialized_start=1174
  _globals['_GENERATERESPONSEFORMAT']._serialized_end=1346
  _globals['_GENERATERESPONSEFORMAT_RESPONSEFORMAT']._serialized_start=1281
  _globals['_GENERATERESPONSEFORMAT_RESPONSEFORMAT']._serialized_end=1335
  _globals['_CONTENT']._serialized_start=1349
  _globals['_CONTENT']._serialized_end=1507
  _globals['_CONTENT_ROLE']._serialized_start=1429
  _globals['_CONTENT_ROLE']._serialized_end=1498
  _globals['_CONTEXT']._serialized_start=1509
  _globals['_CONTEXT']._serialized_end=1549
  _globals['_GENERATEREQUEST']._serialized_start=1552
  _globals['_GENERATEREQUEST']._serialized_end=1767
  _globals['_GENERATIONUSAGE']._serialized_start=1769
  _globals['_GENERATIONUSAGE']._serialized_end=1857
  _globals['_TOKENRESPONSE']._serialized_start=1860
  _globals['_TOKENRESPONSE']._serialized_end=2045
  _globals['_GENERATERESPONSE']._serialized_start=2048
  _globals['_GENERATERESPONSE']._serialized_end=2239
  _globals['_SORTEDEMBEDDING']._serialized_start=2241
  _globals['_SORTEDEMBEDDING']._serialized_end=2287
  _globals['_EMBEDRESPONSE']._serialized_start=2289
  _globals['_EMBEDRESPONSE']._serialized_end=2343
  _globals['_EMBEDREQUEST']._serialized_start=2345
  _globals['_EMBEDREQUEST']._serialized_end=2393
  _globals['_APPUPLOADPROGRESS']._serialized_start=2396
  _globals['_APPUPLOADPROGRESS']._serialized_end=3108
  _globals['_APPUPLOADPROGRESS_APPUPLOADERROR']._serialized_start=2635
  _globals['_APPUPLOADPROGRESS_APPUPLOADERROR']._serialized_end=2670
  _globals['_APPUPLOADPROGRESS_UPLOADSTEP']._serialized_start=2673
  _globals['_APPUPLOADPROGRESS_UPLOADSTEP']._serialized_end=2987
  _globals['_APPUPLOADPROGRESS_UPLOADTYPE']._serialized_start=2989
  _globals['_APPUPLOADPROGRESS_UPLOADTYPE']._serialized_end=3080
  _globals['_TRUFFLESDK']._serialized_start=3343
  _globals['_TRUFFLESDK']._serialized_end=3798
# @@protoc_insertion_point(module_scope)



import typing
from truffle.common import get_logger
from .base import *

logger = get_logger()

def tool_decorator(description: str = None, icon: str = None, predicate: typing.Callable = None):
    def decorator(func):
        assert verify_func(func), f"Function {func.__name__} cannot be a tool"
        func.__truffle_tool__ = True
        func.__truffle_description__ = description
        func.__truffle_icon__ = icon

        func_attrs = ["__truffle_args__", "__truffle_group__", "__truffle_flags__"]
        for attr in func_attrs:
            if not hasattr(func, attr):
                setattr(func, attr, None)

        
        
        if predicate is not None:
            assert verify_predicate(predicate), f"Function {func.__name__} has an invalid predicate"
            func.__truffle_predicate__ = predicate
        logger.debug(f"@truffle.tool({description},{icon}) - {func.__name__} {' - predicate' if predicate else ''}")
        return func
    return decorator

def args_decorator(**kwargs):
    def decorator(func):
        assert verify_arg_descriptions(func.__name__, kwargs)
        func.__truffle_args__ = kwargs

        logger.debug(f"@truffle.args({kwargs}) for function: {func.__name__}")
        return func
    return decorator

def group_decorator(name: str, leader: bool = False): 
    def decorator(func):
        assert verify_func(func), f"Function {func.__name__} cannot be a group"
        func.__truffle_group__ = name
        func.__truffle_group_leader__ = leader
        logger.debug(f"@truffle.group({name}, {leader}) - {func.__name__}")
        return func
    return decorator

def flags_decorator(**kwargs):
    def decorator(func):
        func.__truffle_flags__ = "nonblock" #todo: make list in future 
        logger.debug(f"@truffle.flags({kwargs}) - {func.__name__}")
        return func
    return decorator


import typing
from truffle.common import get_logger
from .decs import tool_decorator, args_decorator, group_decorator, flags_decorator


from .determine_runtime import determine_runtime, RuntimeType

HOST = determine_runtime()


logger = get_logger()

if HOST in [RuntimeType.TRUFFLE, RuntimeType.DEV]:
    from .proprietary import TruffleRuntime
    logger.debug("Using proprietary runtime")
elif HOST == RuntimeType.CLIENT:
    logger.debug("Using public runtime")
    from .public import TruffleClientRuntime as TruffleRuntime
else:
    raise ValueError(f"Invalid runtime type: {HOST}")


def Runtime():
    return TruffleRuntime

def group(name: str, leader: bool = False):
    return group_decorator(name, leader)

def tool(description: str = None, icon: str = None, predicate: typing.Callable = None):
    return tool_decorator(description, icon, predicate)

def flags(**kwargs):
    return flags_decorator(**kwargs)  

def args(**kwargs):
    return args_decorator(**kwargs)
    
def app(**kwargs):
    def decorator(func):
        return func 
    return decorator



from truffle.common import get_logger
from .base import *
from dataclasses import is_dataclass


logger = get_logger()

class TruffleClientRuntime(BaseRuntime):
    def __init__(self):
        self.cls = None

    def build(self, class_instance):
        self.cls = class_instance
        
        # assert tools != {} inside 
        tool_fns = get_truffle_tool_fns(self.cls)

        check_groups(self.cls)

        for name , func in tool_fns.items():
            # must have args and types for args
            args = args_from_func(func)
            assert args != None, f"Function {func.__name__} invalid"
            
            verify_arg_types(name, args) # this can be dome in decs tbh
        
        # maybeeee fuzz every tool ??

        logger.success("Validated all tools, building Truffle app...")
        return self.cls
    




import abc 
import typing
import inspect
import json 
from truffle.common import get_logger
from truffle.types import TruffleFile

class BaseRuntime(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def build(self, class_instance: typing.Any) -> None:
        pass


def is_jsonable(x):
    try:
        json.dumps(x)
        return True
    except (TypeError, OverflowError):
        return False


def get_members(obj : typing.Any, pred: typing.Callable) -> typing.Dict[str, typing.Any]:
    pr = {}
    for name in dir(obj):
        value = getattr(obj, name)
        if not name.startswith('__') and pred(value):
            pr[name] = value
    return pr

def get_non_function_members(obj):
    return get_members(obj, lambda o: not inspect.ismethod(o))

def get_function_members(obj):
    return get_members(obj, inspect.ismethod)
    

def args_from_func(func : typing.Callable) -> typing.Dict[str, typing.Any]:
    type_hints = typing.get_type_hints(func)
    assert callable(func), f"Expected a function, got type: {type(func)} {func}."
    assert type_hints, f"Function {func.__name__} must have type hints."
    
    assert "return" in type_hints, f"Function {func.__name__} must have a return value and type hint."

    args_dict = {}
    for param_name, param in type_hints.items():
        param_type = type_hints.get(param_name, type(None))
        assert param_type.__name__ != "NoneType", (
            f"Function {func.__name__}: Parameter '{param_name}' has no type hint. Make sure to include a type hint."
        )
        args_dict[param_name] = param_type

    assert "return" in args_dict, f"Args dict missing return value for function {func.__name__}."
    return args_dict


def get_truffle_tool_fns(obj):
    tools = {}
    for name, func in get_function_members(obj).items():
        if hasattr(func, "__truffle_tool__"):
            if hasattr(func, "__self__"):
                tools[name] = func.__func__
            else:
                tools[name] = func
                get_logger().warning(f"Function {func.__name__} missing self parameter. Trying to make it work.")
    assert tools != {}, f"Class {obj.__class__.__name__} has no truffle tools defined, don't forget to add @truffle.tool to your functions."
    return tools


def verify_func(func : typing.Callable) -> bool:
    assert len(args_from_func(func)), f"Function {func.__name__} invalid"
    return True

def verify_arg_descriptions(fn_name :str, kwargs : typing.Dict[str, typing.Any]) -> bool:
    assert kwargs != None, f"{fn_name} - truffle.args() requires at least one [name, description] pair, got none"
    assert len(kwargs) > 0, f"{fn_name} - truffle.args() requires at least one [name, description] pair, got none"
    for key, value in kwargs.items():
        assert isinstance(key, str),   f"{fn_name}.args({key}='{value}') - Expected string, got type {type(key)} {key}."
        assert isinstance(value, str), f"{fn_name}.args({key}='{value}') - Expected string, got type {type(value)} {value}."
    return True


def verify_predicate(func : typing.Callable) -> bool:
    assert callable(func), f"Predicate {func.__name__} must be callable"
    assert inspect.isfunction(func), f"Predicate {func.__name__} must be a function"
    # assert no args, except self
    params = [p for p in inspect.signature(func).parameters.values() if p.name != 'self']
    assert len(params) == 0, f"Predicate {func.__name__} must have no arguments (excluding 'self')"
    assert inspect.signature(func).return_annotation == bool, f"Predicate {func.__name__} must return a bool"

    return True


def check_groups(obj : typing.Any) -> bool:
    groups = {}
    for name, func in get_function_members(obj).items():
        if getattr(func, "__truffle_group__", None) is not None:
            is_leader = getattr(func, "__truffle_group_leader__", False)
            group_name = str(func.__truffle_group__)
            if group_name not in groups:
                groups[group_name] = {"leaders": [], "members": []}
            if is_leader:
                groups[group_name]["leaders"].append(func)
            else:
                groups[group_name]["members"].append(func)

    for group_name, group in groups.items():
        assert len(group["leaders"]) > 0, f"Group {group_name} has no leaders, so all tools in the group will be ignored."
        if len(group["members"]) == 0:
            get_logger().warning(f"Group {group_name} has no members, so all tools in the group will be available")

    return True

def verify_arg_types(name: str, args: typing.Dict[str, typing.Any]) -> bool:
    for arg_name, arg_type in args.items():
        
        allowed_arg_types = [str, int, float, bool, list, dict]
        allowed_return_types = [str, int, float, bool, list, dict, TruffleFile]
        
        if arg_name == 'return':
            assert any(arg_type is t for t in allowed_return_types) or hasattr(arg_type, "__origin__"), \
                f"Function {name} return type has invalid type: {arg_type}. Must be one of: {', '.join(t.__name__ for t in allowed_return_types)}"
        else:
            assert any(arg_type is t for t in allowed_arg_types) or hasattr(arg_type, "__origin__"), \
                f"Function {name} argument {arg_name} has invalid type: {arg_type}. Must be one of: {', '.join(t.__name__ for t in allowed_arg_types)}"
        
        if hasattr(arg_type, "__origin__"):
            origin = arg_type.__origin__
            assert origin in (list, dict), f"Function {name} argument {arg_name} uses unsupported generic type: {origin}"
            
            for inner_type in arg_type.__args__:
                assert inner_type in allowed_arg_types and not (hasattr(inner_type, "__origin__") and inner_type.__origin__ is typing.Union), \
                f"Function {name} argument {arg_name} has invalid inner type: {inner_type}. Optional types are not supported."


import os
from datetime import datetime
from dataclasses import dataclass

@dataclass
class Colors:
    MAIN: str = "\033[38;2;95;173;235m"      # #5FADEB - Primary actions/success
    SECONDARY: str = "\033[38;2;74;155;217m"  # #4A9BD9 - Info/progress
    DIM_GRAY: str = "\033[38;2;128;128;128m"  # #808080 - Technical details
    ERROR: str = "\033[38;2;255;59;48m"       # #FF3B30 - Errors
    WARNING: str = "\033[38;2;255;149;0m"     # #FF9500 - Warnings
    RESET: str = "\033[0m"

@dataclass
class Symbols:
    SUCCESS: str = "✓"
    ERROR: str = "✗"
    WARNING: str = "!"
    INFO: str = "ℹ"
    PROMPT: str = "❯"

class Logger:
    def __init__(self) -> None:
        pass

    def success(self, message: str, end: str = "\n") -> None:
        print(f"{Colors.MAIN}{Symbols.SUCCESS} {message}{Colors.RESET}", end=end)

    def error(self, message: str, end: str = "\n") -> None:
        print(f"{Colors.ERROR}{Symbols.ERROR} {message}{Colors.RESET}", end=end)

    def warning(self, message: str, end: str = "\n") -> None:
        print(f"{Colors.WARNING}{Symbols.WARNING} {message}{Colors.RESET}", end=end)
    
    def info(self, message: str, end: str = "\n") -> None:
        print(f"{Colors.SECONDARY}{Symbols.INFO} {message}{Colors.RESET}", end=end)
    
    def plain_info(self, message: str, end: str = "\n") -> None:
        print(f"{Colors.SECONDARY}{message}{Colors.RESET}", end=end)

    def input(self, message: str) -> str:
        return input(f"{Colors.SECONDARY}{Symbols.PROMPT} {message}{Colors.RESET}")

    def debug(self, message: str, end: str = "\n") -> None:
        if os.getenv("TRUFFLE_DEBUG"):
            timestamp = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
            print(f"{Colors.DIM_GRAY}[{timestamp}] {Symbols.INFO} {message}{Colors.RESET}", end=end)


from setuptools import setup, find_packages

setup(
    name="hyphae",
    version="0.9.5",
    description="The framework for developing TruffleOS agentic applications",
    author="Deepshard",
    author_email="accounts@deepshard.org",
    license="MIT",
    packages=find_packages(),
    install_requires=[
        "grpcio==1.69.0",
        "grpcio-reflection==1.69.0",
        "protobuf==5.29.3",
        "requests>=2.32.3"
    ],
    python_requires=">=3.10",
    setup_requires=["wheel"],
    entry_points={
        "console_scripts": [
            "truffle=truffle.cli:main",
        ],
    },      
)



from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class GenerateFinishReason(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    FINISH_REASON_UNSPECIFIED: _ClassVar[GenerateFinishReason]
    FINISH_REASON_LENGTH: _ClassVar[GenerateFinishReason]
    FINISH_REASON_STOP: _ClassVar[GenerateFinishReason]
    FINISH_REASON_ERROR: _ClassVar[GenerateFinishReason]
    FINISH_REASON_USER: _ClassVar[GenerateFinishReason]

class TruffleType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    TRUFFLE_UNSPECIFIED: _ClassVar[TruffleType]
    TRUFFLE_IMAGE: _ClassVar[TruffleType]
    TRUFFLE_FILE: _ClassVar[TruffleType]
FINISH_REASON_UNSPECIFIED: GenerateFinishReason
FINISH_REASON_LENGTH: GenerateFinishReason
FINISH_REASON_STOP: GenerateFinishReason
FINISH_REASON_ERROR: GenerateFinishReason
FINISH_REASON_USER: GenerateFinishReason
TRUFFLE_UNSPECIFIED: TruffleType
TRUFFLE_IMAGE: TruffleType
TRUFFLE_FILE: TruffleType

class ModelDescription(_message.Message):
    __slots__ = ("model_id", "name", "description", "capabilities", "type", "is_local")
    class ModelType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        MODEL_UNSPECIFIED: _ClassVar[ModelDescription.ModelType]
        MODEL_SMART: _ClassVar[ModelDescription.ModelType]
        MODEL_FAST: _ClassVar[ModelDescription.ModelType]
        MODEL_VISION: _ClassVar[ModelDescription.ModelType]
        MODEL_AGI: _ClassVar[ModelDescription.ModelType]
    MODEL_UNSPECIFIED: ModelDescription.ModelType
    MODEL_SMART: ModelDescription.ModelType
    MODEL_FAST: ModelDescription.ModelType
    MODEL_VISION: ModelDescription.ModelType
    MODEL_AGI: ModelDescription.ModelType
    class Capabilities(_message.Message):
        __slots__ = ("structured_output", "image_support", "reasoner", "world_domination", "languages")
        STRUCTURED_OUTPUT_FIELD_NUMBER: _ClassVar[int]
        IMAGE_SUPPORT_FIELD_NUMBER: _ClassVar[int]
        REASONER_FIELD_NUMBER: _ClassVar[int]
        WORLD_DOMINATION_FIELD_NUMBER: _ClassVar[int]
        LANGUAGES_FIELD_NUMBER: _ClassVar[int]
        structured_output: bool
        image_support: bool
        reasoner: bool
        world_domination: bool
        languages: _containers.RepeatedScalarFieldContainer[str]
        def __init__(self, structured_output: bool = ..., image_support: bool = ..., reasoner: bool = ..., world_domination: bool = ..., languages: _Optional[_Iterable[str]] = ...) -> None: ...
    MODEL_ID_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    CAPABILITIES_FIELD_NUMBER: _ClassVar[int]
    TYPE_FIELD_NUMBER: _ClassVar[int]
    IS_LOCAL_FIELD_NUMBER: _ClassVar[int]
    model_id: int
    name: str
    description: str
    capabilities: ModelDescription.Capabilities
    type: ModelDescription.ModelType
    is_local: bool
    def __init__(self, model_id: _Optional[int] = ..., name: _Optional[str] = ..., description: _Optional[str] = ..., capabilities: _Optional[_Union[ModelDescription.Capabilities, _Mapping]] = ..., type: _Optional[_Union[ModelDescription.ModelType, str]] = ..., is_local: bool = ...) -> None: ...

class GetModelsRequest(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class GetModelsResponse(_message.Message):
    __slots__ = ("models",)
    MODELS_FIELD_NUMBER: _ClassVar[int]
    models: _containers.RepeatedCompositeFieldContainer[ModelDescription]
    def __init__(self, models: _Optional[_Iterable[_Union[ModelDescription, _Mapping]]] = ...) -> None: ...

class GlobalsRequest(_message.Message):
    __slots__ = ("key", "value")
    KEY_FIELD_NUMBER: _ClassVar[int]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    key: str
    value: str
    def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...

class GlobalsResponse(_message.Message):
    __slots__ = ("value",)
    VALUE_FIELD_NUMBER: _ClassVar[int]
    value: str
    def __init__(self, value: _Optional[str] = ...) -> None: ...

class SystemToolRequest(_message.Message):
    __slots__ = ("tool_name", "args")
    class ArgsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    TOOL_NAME_FIELD_NUMBER: _ClassVar[int]
    ARGS_FIELD_NUMBER: _ClassVar[int]
    tool_name: str
    args: _containers.ScalarMap[str, str]
    def __init__(self, tool_name: _Optional[str] = ..., args: _Optional[_Mapping[str, str]] = ...) -> None: ...

class SystemToolResponse(_message.Message):
    __slots__ = ("response", "results", "error")
    class ResultsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    RESPONSE_FIELD_NUMBER: _ClassVar[int]
    RESULTS_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    response: str
    results: _containers.ScalarMap[str, str]
    error: str
    def __init__(self, response: _Optional[str] = ..., results: _Optional[_Mapping[str, str]] = ..., error: _Optional[str] = ...) -> None: ...

class SDKResponse(_message.Message):
    __slots__ = ("ok", "error")
    OK_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    ok: bool
    error: str
    def __init__(self, ok: bool = ..., error: _Optional[str] = ...) -> None: ...

class ToolUpdateRequest(_message.Message):
    __slots__ = ("friendly_description",)
    FRIENDLY_DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    friendly_description: str
    def __init__(self, friendly_description: _Optional[str] = ...) -> None: ...

class UserRequest(_message.Message):
    __slots__ = ("message", "reason")
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    REASON_FIELD_NUMBER: _ClassVar[int]
    message: str
    reason: str
    def __init__(self, message: _Optional[str] = ..., reason: _Optional[str] = ...) -> None: ...

class UserResponse(_message.Message):
    __slots__ = ("response", "error", "attached_files")
    RESPONSE_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    ATTACHED_FILES_FIELD_NUMBER: _ClassVar[int]
    response: str
    error: str
    attached_files: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, response: _Optional[str] = ..., error: _Optional[str] = ..., attached_files: _Optional[_Iterable[str]] = ...) -> None: ...

class GenerateResponseFormat(_message.Message):
    __slots__ = ("format", "schema")
    class ResponseFormat(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        RESPONSE_TEXT: _ClassVar[GenerateResponseFormat.ResponseFormat]
        RESPONSE_JSON: _ClassVar[GenerateResponseFormat.ResponseFormat]
    RESPONSE_TEXT: GenerateResponseFormat.ResponseFormat
    RESPONSE_JSON: GenerateResponseFormat.ResponseFormat
    FORMAT_FIELD_NUMBER: _ClassVar[int]
    SCHEMA_FIELD_NUMBER: _ClassVar[int]
    format: GenerateResponseFormat.ResponseFormat
    schema: str
    def __init__(self, format: _Optional[_Union[GenerateResponseFormat.ResponseFormat, str]] = ..., schema: _Optional[str] = ...) -> None: ...

class Content(_message.Message):
    __slots__ = ("role", "content", "data")
    class Role(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        ROLE_INVALID: _ClassVar[Content.Role]
        ROLE_USER: _ClassVar[Content.Role]
        ROLE_AI: _ClassVar[Content.Role]
        ROLE_SYSTEM: _ClassVar[Content.Role]
    ROLE_INVALID: Content.Role
    ROLE_USER: Content.Role
    ROLE_AI: Content.Role
    ROLE_SYSTEM: Content.Role
    ROLE_FIELD_NUMBER: _ClassVar[int]
    CONTENT_FIELD_NUMBER: _ClassVar[int]
    DATA_FIELD_NUMBER: _ClassVar[int]
    role: Content.Role
    content: str
    data: bytes
    def __init__(self, role: _Optional[_Union[Content.Role, str]] = ..., content: _Optional[str] = ..., data: _Optional[bytes] = ...) -> None: ...

class Context(_message.Message):
    __slots__ = ("history",)
    HISTORY_FIELD_NUMBER: _ClassVar[int]
    history: _containers.RepeatedCompositeFieldContainer[Content]
    def __init__(self, history: _Optional[_Iterable[_Union[Content, _Mapping]]] = ...) -> None: ...

class GenerateRequest(_message.Message):
    __slots__ = ("message", "context", "max_tokens", "model_id", "temperature", "fmt")
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    CONTEXT_FIELD_NUMBER: _ClassVar[int]
    MAX_TOKENS_FIELD_NUMBER: _ClassVar[int]
    MODEL_ID_FIELD_NUMBER: _ClassVar[int]
    TEMPERATURE_FIELD_NUMBER: _ClassVar[int]
    FMT_FIELD_NUMBER: _ClassVar[int]
    message: str
    context: Context
    max_tokens: int
    model_id: int
    temperature: float
    fmt: GenerateResponseFormat
    def __init__(self, message: _Optional[str] = ..., context: _Optional[_Union[Context, _Mapping]] = ..., max_tokens: _Optional[int] = ..., model_id: _Optional[int] = ..., temperature: _Optional[float] = ..., fmt: _Optional[_Union[GenerateResponseFormat, _Mapping]] = ...) -> None: ...

class GenerationUsage(_message.Message):
    __slots__ = ("prompt_tokens", "completion_tokens", "approx_time")
    PROMPT_TOKENS_FIELD_NUMBER: _ClassVar[int]
    COMPLETION_TOKENS_FIELD_NUMBER: _ClassVar[int]
    APPROX_TIME_FIELD_NUMBER: _ClassVar[int]
    prompt_tokens: int
    completion_tokens: int
    approx_time: int
    def __init__(self, prompt_tokens: _Optional[int] = ..., completion_tokens: _Optional[int] = ..., approx_time: _Optional[int] = ...) -> None: ...

class TokenResponse(_message.Message):
    __slots__ = ("token", "finish_reason", "usage", "error")
    TOKEN_FIELD_NUMBER: _ClassVar[int]
    FINISH_REASON_FIELD_NUMBER: _ClassVar[int]
    USAGE_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    token: str
    finish_reason: GenerateFinishReason
    usage: GenerationUsage
    error: str
    def __init__(self, token: _Optional[str] = ..., finish_reason: _Optional[_Union[GenerateFinishReason, str]] = ..., usage: _Optional[_Union[GenerationUsage, _Mapping]] = ..., error: _Optional[str] = ...) -> None: ...

class GenerateResponse(_message.Message):
    __slots__ = ("response", "finish_reason", "usage", "error")
    RESPONSE_FIELD_NUMBER: _ClassVar[int]
    FINISH_REASON_FIELD_NUMBER: _ClassVar[int]
    USAGE_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    response: str
    finish_reason: GenerateFinishReason
    usage: GenerationUsage
    error: str
    def __init__(self, response: _Optional[str] = ..., finish_reason: _Optional[_Union[GenerateFinishReason, str]] = ..., usage: _Optional[_Union[GenerationUsage, _Mapping]] = ..., error: _Optional[str] = ...) -> None: ...

class SortedEmbedding(_message.Message):
    __slots__ = ("text", "score")
    TEXT_FIELD_NUMBER: _ClassVar[int]
    SCORE_FIELD_NUMBER: _ClassVar[int]
    text: str
    score: float
    def __init__(self, text: _Optional[str] = ..., score: _Optional[float] = ...) -> None: ...

class EmbedResponse(_message.Message):
    __slots__ = ("results",)
    RESULTS_FIELD_NUMBER: _ClassVar[int]
    results: _containers.RepeatedCompositeFieldContainer[SortedEmbedding]
    def __init__(self, results: _Optional[_Iterable[_Union[SortedEmbedding, _Mapping]]] = ...) -> None: ...

class EmbedRequest(_message.Message):
    __slots__ = ("documents", "query")
    DOCUMENTS_FIELD_NUMBER: _ClassVar[int]
    QUERY_FIELD_NUMBER: _ClassVar[int]
    documents: _containers.RepeatedScalarFieldContainer[str]
    query: str
    def __init__(self, documents: _Optional[_Iterable[str]] = ..., query: _Optional[str] = ...) -> None: ...

class AppUploadProgress(_message.Message):
    __slots__ = ("step", "progress", "type", "latest_logs", "substep", "error")
    class UploadStep(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        STEP_UNKNOWN: _ClassVar[AppUploadProgress.UploadStep]
        STEP_UPLOAD_BUNDLE: _ClassVar[AppUploadProgress.UploadStep]
        STEP_VERIFY_BUNDLE: _ClassVar[AppUploadProgress.UploadStep]
        STEP_SYNC_WITH_SERVER: _ClassVar[AppUploadProgress.UploadStep]
        STEP_WRITE_BUNDLE: _ClassVar[AppUploadProgress.UploadStep]
        STEP_CREATE_CONTAINER: _ClassVar[AppUploadProgress.UploadStep]
        STEP_INSTALL_DEPENDENCIES: _ClassVar[AppUploadProgress.UploadStep]
        STEP_GENERATE_CLASSIFIER_DATA: _ClassVar[AppUploadProgress.UploadStep]
        STEP_SAVE_CONTAINER: _ClassVar[AppUploadProgress.UploadStep]
        STEP_GATHER_TOOLS: _ClassVar[AppUploadProgress.UploadStep]
        STEP_VERIFY_INSTALL: _ClassVar[AppUploadProgress.UploadStep]
        STEP_INSTALLED: _ClassVar[AppUploadProgress.UploadStep]
    STEP_UNKNOWN: AppUploadProgress.UploadStep
    STEP_UPLOAD_BUNDLE: AppUploadProgress.UploadStep
    STEP_VERIFY_BUNDLE: AppUploadProgress.UploadStep
    STEP_SYNC_WITH_SERVER: AppUploadProgress.UploadStep
    STEP_WRITE_BUNDLE: AppUploadProgress.UploadStep
    STEP_CREATE_CONTAINER: AppUploadProgress.UploadStep
    STEP_INSTALL_DEPENDENCIES: AppUploadProgress.UploadStep
    STEP_GENERATE_CLASSIFIER_DATA: AppUploadProgress.UploadStep
    STEP_SAVE_CONTAINER: AppUploadProgress.UploadStep
    STEP_GATHER_TOOLS: AppUploadProgress.UploadStep
    STEP_VERIFY_INSTALL: AppUploadProgress.UploadStep
    STEP_INSTALLED: AppUploadProgress.UploadStep
    class UploadType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        UPLOAD_TYPE_UNDETERMINED: _ClassVar[AppUploadProgress.UploadType]
        UPLOAD_TYPE_INSTALL: _ClassVar[AppUploadProgress.UploadType]
        UPLOAD_TYPE_UPDATE: _ClassVar[AppUploadProgress.UploadType]
    UPLOAD_TYPE_UNDETERMINED: AppUploadProgress.UploadType
    UPLOAD_TYPE_INSTALL: AppUploadProgress.UploadType
    UPLOAD_TYPE_UPDATE: AppUploadProgress.UploadType
    class AppUploadError(_message.Message):
        __slots__ = ("raw_error",)
        RAW_ERROR_FIELD_NUMBER: _ClassVar[int]
        raw_error: str
        def __init__(self, raw_error: _Optional[str] = ...) -> None: ...
    STEP_FIELD_NUMBER: _ClassVar[int]
    PROGRESS_FIELD_NUMBER: _ClassVar[int]
    TYPE_FIELD_NUMBER: _ClassVar[int]
    LATEST_LOGS_FIELD_NUMBER: _ClassVar[int]
    SUBSTEP_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    step: AppUploadProgress.UploadStep
    progress: int
    type: AppUploadProgress.UploadType
    latest_logs: str
    substep: str
    error: AppUploadProgress.AppUploadError
    def __init__(self, step: _Optional[_Union[AppUploadProgress.UploadStep, str]] = ..., progress: _Optional[int] = ..., type: _Optional[_Union[AppUploadProgress.UploadType, str]] = ..., latest_logs: _Optional[str] = ..., substep: _Optional[str] = ..., error: _Optional[_Union[AppUploadProgress.AppUploadError, _Mapping]] = ...) -> None: ...
